import claripy
import pickle

import logging

l = logging.getLogger("claripy.test.serial")


def test_pickle_ast():
    bz = claripy.backends.z3

    a = claripy.BVV(1, 32)
    b = claripy.BVS("x", 32, explicit_name=True)

    c = a + b
    assert bz.convert(c).__module__ == "z3.z3"
    assert str(bz.convert(c)), "1 + x"

    c_copy = pickle.loads(pickle.dumps(c, -1))
    assert c_copy is c
    assert bz.convert(c_copy).__module__ == "z3.z3"
    assert str(bz.convert(c_copy)) == "1 + x"


def test_pickle_frontend():
    s = claripy.Solver()
    x = claripy.BVS("x", 32)

    s.add(x == 1)
    assert s.eval(x, 10), (1,)

    ss = pickle.dumps(s)
    del s
    import gc

    gc.collect()

    s = pickle.loads(ss)
    assert s.eval(x, 10), (1,)


def test_identity():
    l.info("Running test_identity")

    a = claripy.BVV(1, 32)
    b = claripy.BVS("x", 32)
    c = a + b
    d = (
        a
        + b
        + b
        + b
        + b
        + b
        + b
        + b
        + b
        + b
        + b
        + b
        + b
        + b
        + b
        + b
        + b
        + b
        + b
        + b
        + b
        + b
        + b
        + b
        + b
        + b
        + b
        + b
        + b
        + b
        + b
        + b
        + b
        + b
        + b
        + b
        + b
        + b
        + b
        + b
        + b
        + b
        + b
        + b
        + b
        + b
        + b
        + b
        + b
        + b
        + b
        + b
        + b
    )

    l.debug("Storing!")
    c_info = pickle.dumps(c)
    d_info = pickle.dumps(d)

    cc = pickle.loads(c_info)
    assert str(cc) == str(c)
    cd = pickle.loads(d_info)
    assert str(cd) == str(d)
    assert c.args[0] is d.args[0]

    l.debug("Time to test some solvers!")
    s = claripy.Solver()
    x = claripy.BVS("x", 32)
    s.add(x == 3)
    s.finalize()
    ss = pickle.loads(pickle.dumps(s))
    assert str(s.constraints) == str(ss.constraints)
    assert str(s.variables) == str(ss.variables)

    s = claripy.SolverComposite()
    x = claripy.BVS("x", 32)
    s.add(x == 3)
    s.finalize()
    ss = pickle.loads(pickle.dumps(s))
    old_constraint_sets = [[hash(j) for j in k.constraints] for k in s._solver_list]
    new_constraint_sets = [[hash(j) for j in k.constraints] for k in ss._solver_list]
    assert old_constraint_sets == new_constraint_sets
    assert str(s.variables) == str(ss.variables)


if __name__ == "__main__":
    test_pickle_ast()
    test_pickle_frontend()
    test_identity()



# to enable debug mode, please set _DEBUG to True via the set_debug method.
_DEBUG = True


def set_debug(enabled):
    """
    Enable or disable the debug mode. In debug mode, a bunch of extra checks in claripy will be executed. You'll want to
    disable debug mode if you are running performance critical code.
    """

    global _DEBUG
    _DEBUG = enabled



import claripy


def test_complex_guy():
    guy_wide = claripy.widen(
        claripy.union(
            claripy.union(
                claripy.union(claripy.BVV(0, 32), claripy.BVV(1, 32)),
                claripy.union(claripy.BVV(0, 32), claripy.BVV(1, 32)) + claripy.BVV(1, 32),
            ),
            claripy.union(
                claripy.union(claripy.BVV(0, 32), claripy.BVV(1, 32)),
                claripy.union(claripy.BVV(0, 32), claripy.BVV(1, 32)) + claripy.BVV(1, 32),
            )
            + claripy.BVV(1, 32),
        ),
        claripy.union(
            claripy.union(
                claripy.union(
                    claripy.union(claripy.BVV(0, 32), claripy.BVV(1, 32)),
                    claripy.union(claripy.BVV(0, 32), claripy.BVV(1, 32)) + claripy.BVV(1, 32),
                ),
                claripy.union(
                    claripy.union(claripy.BVV(0, 32), claripy.BVV(1, 32)),
                    claripy.union(claripy.BVV(0, 32), claripy.BVV(1, 32)) + claripy.BVV(1, 32),
                )
                + claripy.BVV(1, 32),
            ),
            claripy.union(
                claripy.union(
                    claripy.union(claripy.BVV(0, 32), claripy.BVV(1, 32)),
                    claripy.union(claripy.BVV(0, 32), claripy.BVV(1, 32)) + claripy.BVV(1, 32),
                ),
                claripy.union(
                    claripy.union(claripy.BVV(0, 32), claripy.BVV(1, 32)),
                    claripy.union(claripy.BVV(0, 32), claripy.BVV(1, 32)) + claripy.BVV(1, 32),
                )
                + claripy.BVV(1, 32),
            )
            + claripy.BVV(1, 32),
        ),
    )
    guy_inc = guy_wide + claripy.BVV(1, 32)
    guy_zx = claripy.ZeroExt(32, guy_inc)

    s, r = claripy.balancer.Balancer(claripy.backends.vsa, guy_inc <= claripy.BVV(39, 32)).compat_ret
    assert s
    assert r[0][0] is guy_wide
    assert claripy.backends.vsa.min(r[0][1]) == 0
    assert set(claripy.backends.vsa.eval(r[0][1], 1000)) == set([4294967295] + list(range(39)))

    s, r = claripy.balancer.Balancer(claripy.backends.vsa, guy_zx <= claripy.BVV(39, 64)).compat_ret
    assert r[0][0] is guy_wide
    assert claripy.backends.vsa.min(r[0][1]) == 0
    assert set(claripy.backends.vsa.eval(r[0][1], 1000)) == set([4294967295] + list(range(39)))


def test_simple_guy():
    x = claripy.BVS("x", 32)
    s, r = claripy.balancer.Balancer(claripy.backends.vsa, x <= claripy.BVV(39, 32)).compat_ret
    assert s
    assert r[0][0] is x
    assert claripy.backends.vsa.min(r[0][1]) == 0
    assert claripy.backends.vsa.max(r[0][1]) == 39

    s, r = claripy.balancer.Balancer(claripy.backends.vsa, x + 1 <= claripy.BVV(39, 32)).compat_ret
    assert s
    assert r[0][0] is x
    all_vals = r[0][1]._model_vsa.eval(1000)
    assert len(all_vals)
    assert min(all_vals) == 0
    assert max(all_vals) == 4294967295
    all_vals.remove(4294967295)
    assert max(all_vals) == 38


def test_widened_guy():
    w = claripy.widen(claripy.BVV(1, 32), claripy.BVV(0, 32))
    s, r = claripy.balancer.Balancer(claripy.backends.vsa, w <= claripy.BVV(39, 32)).compat_ret
    assert s
    assert r[0][0] is w
    assert claripy.backends.vsa.min(r[0][1]) == 0
    assert claripy.backends.vsa.max(r[0][1]) == 1  # used to be 39, but that was a bug in the VSA widening

    s, r = claripy.balancer.Balancer(claripy.backends.vsa, w + 1 <= claripy.BVV(39, 32)).compat_ret
    assert s
    assert r[0][0] is w
    assert claripy.backends.vsa.min(r[0][1]) == 0
    all_vals = r[0][1]._model_vsa.eval(1000)
    assert set(all_vals) == {4294967295, 0, 1}


def test_overflow():
    x = claripy.BVS("x", 32)

    print("x + 10 <= 20")
    s, r = claripy.balancer.Balancer(claripy.backends.vsa, x + 10 <= claripy.BVV(20, 32)).compat_ret
    # mn,mx = claripy.backends.vsa.min(r[0][1]), claripy.backends.vsa.max(r[0][1])
    assert s
    assert r[0][0] is x
    assert set(claripy.backends.vsa.eval(r[0][1], 1000)) == {
        4294967286,
        4294967287,
        4294967288,
        4294967289,
        4294967290,
        4294967291,
        4294967292,
        4294967293,
        4294967294,
        4294967295,
        0,
        1,
        2,
        3,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
    }

    # print("0 <= x + 10")
    # s,r = claripy.balancer.Balancer(claripy.backends.vsa, 0 <= x + 10).compat_ret
    # assert s
    # assert r[0][0] is x

    print("x - 10 <= 20")
    s, r = claripy.balancer.Balancer(claripy.backends.vsa, x - 10 <= claripy.BVV(20, 32)).compat_ret
    assert s
    assert r[0][0] is x
    assert set(claripy.backends.vsa.eval(r[0][1], 1000)) == set(range(10, 31))

    # print("0 <= x - 10")
    # s,r = claripy.balancer.Balancer(claripy.backends.vsa, 0 <= x - 10).compat_ret
    # assert s
    # assert r[0][0] is x


def test_extract_zeroext():
    x = claripy.BVS("x", 8)
    expr = claripy.Extract(31, 0, claripy.ZeroExt(56, x)) <= claripy.BVV(0xE, 32)
    s, r = claripy.balancer.Balancer(claripy.backends.vsa, expr).compat_ret

    assert s is True
    assert len(r) == 1
    assert r[0][0] is x


def test_complex_case_0():
    #

    """
    <Bool ((0#48 .. Reverse(unconstrained_read_69_16)) << 0x30) <= 0x40000000000000>

    Created by VEX running on the following x86_64 assembly:
    cmp     word ptr [rdi], 40h
    ja      skip
    """

    x = claripy.BVS("x", 16)
    expr = (claripy.ZeroExt(48, claripy.Reverse(x)) << 0x30) <= 0x40000000000000

    s, r = claripy.balancer.Balancer(claripy.backends.vsa, expr).compat_ret

    assert s
    assert r[0][0] is x
    assert set(claripy.backends.vsa.eval(r[0][1], 1000)) == set(range(0, 65 * 0x100, 0x100))


def test_complex_case_1():
    #

    """
    <Bool (0#31 .. (if (0x8 < (0#32 .. (0xfffffffe + reg_284_0_32{UNINITIALIZED}))) then 1 else 0)) == 0x0>

    Created by VEX running on the following S390X assembly:
    0x40062c:       ahik    %r2, %r11, -2
    0x400632:       clijh   %r2, 8, 0x40065c

    IRSB {
       t0:Ity_I32 t1:Ity_I32 t2:Ity_I32 t3:Ity_I32 t4:Ity_I32 t5:Ity_I32 t6:Ity_I64 t7:Ity_I64 t8:Ity_I64 t9:Ity_I64 t10:Ity_I32 t11:Ity_I1 t12:Ity_I64 t13:Ity_I64 t14:Ity_I64 t15:Ity_I32 t16:Ity_I1

       00 | ------ IMark(0x40062c, 6, 0) ------
       01 | t0 = GET:I32(r11_32)
       02 | t1 = Add32(0xfffffffe,t0)
       03 | PUT(352) = 0x0000000000000003
       04 | PUT(360) = 0xfffffffffffffffe
       05 | t13 = 32Sto64(t0)
       06 | t7 = t13
       07 | PUT(368) = t7
       08 | PUT(376) = 0x0000000000000000
       09 | PUT(r2_32) = t1
       10 | PUT(ia) = 0x0000000000400632
       11 | ------ IMark(0x400632, 6, 0) ------
       12 | t14 = 32Uto64(t1)
       13 | t8 = t14
       14 | t16 = CmpLT64U(0x0000000000000008,t8)
       15 | t15 = 1Uto32(t16)
       16 | t10 = t15
       17 | t11 = CmpNE32(t10,0x00000000)
       18 | if (t11) { PUT(ia) = 0x40065c; Ijk_Boring }
       NEXT: PUT(ia) = 0x0000000000400638; Ijk_Boring
    }
    """

    x = claripy.BVS("x", 32)
    expr = claripy.ZeroExt(
        31, claripy.If(claripy.BVV(0x8, 32) < claripy.BVV(0xFFFFFFFE, 32) + x, claripy.BVV(1, 1), claripy.BVV(0, 1))
    ) == claripy.BVV(0, 32)
    s, r = claripy.balancer.Balancer(claripy.backends.vsa, expr).compat_ret

    assert s is True
    assert len(r) == 1
    assert r[0][0] is x


def test_complex_case_2():
    x = claripy.BVS("x", 32)
    expr = claripy.ZeroExt(
        31, claripy.If(claripy.BVV(0xC, 32) < x, claripy.BVV(1, 1), claripy.BVV(0, 1))
    ) == claripy.BVV(0, 32)
    s, r = claripy.balancer.Balancer(claripy.backends.vsa, expr).compat_ret

    assert s is True
    assert len(r) == 1
    assert r[0][0] is x


if __name__ == "__main__":
    test_overflow()
    test_simple_guy()
    test_widened_guy()
    test_complex_guy()
    test_complex_case_0()
    test_complex_case_1()
    test_complex_case_2()
    test_extract_zeroext()



import claripy


def test_lite_repr():
    one = claripy.BVV(1, 8)
    two = claripy.BVV(2, 8)
    a = claripy.BVS("a", 8, explicit_name=True)
    b = claripy.BVS("b", 8, explicit_name=True)

    assert (a + one * b + two).shallow_repr() == "<BV8 a + 1 * b + 2>"
    assert ((a + one) * (b + two)).shallow_repr() == "<BV8 (a + 1) * (b + 2)>"
    assert (a * one + b * two).shallow_repr() == "<BV8 a * 1 + b * 2>"
    assert (
        (one + a) * (two + b) + (two + a) * (one + b)
    ).shallow_repr() == "<BV8 (1 + a) * (2 + b) + (2 + a) * (1 + b)>"


def test_associativity():
    x = claripy.BVS("x", 8, explicit_name=True)
    y = claripy.BVS("y", 8, explicit_name=True)
    z = claripy.BVS("z", 8, explicit_name=True)
    w = claripy.BVS("w", 8, explicit_name=True)

    assert (x - (y - (z - w))).shallow_repr() == "<BV8 x - (y - (z - w))>"
    assert (x - y - z - w).shallow_repr() == "<BV8 x - y - z - w>"
    assert (x * (y * (z * w))).shallow_repr() == (x * y * z * w).shallow_repr()
    assert (x * y * z * w).shallow_repr() == "<BV8 x * y * z * w>"
    assert (x + y - z - w).shallow_repr() == "<BV8 x + y - z - w>"
    assert (x + (y - (z - w))).shallow_repr() == "<BV8 x + (y - (z - w))>"
    assert (x * y / z % w).shallow_repr() == "<BV8 x * y / z % w>"
    assert (x * (y / (z % w))).shallow_repr() == "<BV8 x * (y / (z % w))>"


if __name__ == "__main__":
    test_lite_repr()
    test_associativity()



import claripy

import logging

l = logging.getLogger("claripy.test.replacements")


def test_replacement_solver():
    sr = claripy.SolverReplacement(claripy.SolverVSA(), replace_constraints=True, complex_auto_replace=True)
    x = claripy.BVS("x", 32)

    sr.add(x + 8 == 10)
    assert sr.max(x) == sr.min(x)

    sr2 = sr.branch()
    sr2.add(x + 8 < 2000)
    assert sr2.max(x) == sr2.min(x) == sr.max(x)


def test_contradiction():
    sr = claripy.SolverReplacement(claripy.Solver(), replace_constraints=True)
    x = claripy.BVS("x", 32)

    sr.add(x == 10)
    assert sr.satisfiable()
    assert sr.eval(x, 10) == (10,)

    sr.add(x == 100)
    assert not sr.satisfiable()


def test_branching_replacement_solver():
    #
    # Simple case: replaceable thing first
    #

    x = claripy.BVS("x", 32)
    s0 = claripy.SolverReplacement(claripy.Solver())
    s0.add(x == 0)

    s1a = s0.branch()
    s1b = s0.branch()

    s1a.add(x == 0)
    s1b.add(x != 0)

    assert s1a.satisfiable()
    assert not s1b.satisfiable()

    #
    # Slightly more complex: different ==
    #

    x = claripy.BVS("x", 32)
    s0 = claripy.SolverReplacement(claripy.Solver())
    s0.add(x == 0)

    s1a = s0.branch()
    s1b = s0.branch()

    s1a.add(x == 0)
    s1b.add(x == 1)

    assert s1a.satisfiable()
    assert not s1b.satisfiable()

    #
    # Complex case: non-replaceable thing first
    #

    # x = claripy.BVS('x', 32)
    # s0 = claripy.SolverReplacement(claripy.Solver())
    # s0.add(x != 0)
    # s1a = s0.branch()
    # s1b = s0.branch()
    # s1a.add(x != 0)
    # s1b.add(x == 0)
    # assert s1a.satisfiable()
    # assert not s1b.satisfiable()


if __name__ == "__main__":
    test_branching_replacement_solver()
    test_replacement_solver()
    test_contradiction()



import subprocess
import logging

import re
from . import SMTLibSolverBackend, PopenSolverProxy
from ...errors import MissingSolverError

log = logging.getLogger(__name__)


def get_version():
    try:
        version_string = subprocess.check_output(["abc", "--help"]).decode("utf-8")
        # version_match = re.match('This is CVC4 version (.*)\n', version_string)

        # if not version_match:
        #    return False, None, "Found malformed version string: {}".format(version_string)

        return True, None, None  # True, version_match.group(1), None

    except subprocess.CalledProcessError as ex:
        return False, None, f"Not found, error: {ex}"
    except OSError as ex:
        return False, None, f"Not found, error: {ex}"


IS_INSTALLED, VERSION, ERROR = get_version()


class ABCProxy(PopenSolverProxy):
    def __init__(self):
        self.installed = False
        p = None
        super().__init__(p)

    def create_process(self):
        if not IS_INSTALLED:
            raise MissingSolverError("ABC not found! Please install ABC before using this backend")
        p = subprocess.Popen(["abc"], stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        self.installed = True
        return p


class SolverBackendABC(SMTLibSolverBackend):
    def solver(self, timeout=None, max_memory=None):
        """
        This function should return an instance of whatever object handles
        solving for this backend. For example, in Z3, this would be z3.Solver().
        """
        return ABCProxy()


from ... import backend_manager as backend_manager

backend_manager.backends._register_backend(SolverBackendABC(), "smtlib_abc", False, False)



import claripy


def test_issue16():
    s = claripy.SolverComposite()

    c = claripy.BVS("test", 32)
    s.add(c[7:0] != 0)

    assert s.satisfiable()
    s.add(c == 0)

    # print(s.satisfiable())
    assert not s.satisfiable(extra_constraints=[claripy.BVS("lol", 32) == 0])
    assert not s.satisfiable()


if __name__ == "__main__":
    test_issue16()
